/* 

    umesh2dgui, a GUI for umesh2d grid generator 

    Copyright (C) 2003 Gabriele Dini Ciacci
  
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "u2d_core.h"

bool isnear(double toll, double x, double y){

  if(fabs(x - y)  > toll) return false;
  else return true;

}

bool isnear(double toll, double x1, double y1, double x2, double y2){

  if(hypot(x1-x2,y1-y2) > toll) return false;
  else return true;

}

bool isnear(double toll, Primitive *primX, Primitive *primY){
  /* for now do not use this */
  /* we must check 4 combinations of points and first I have to
   * figure out witch is better way to do this*/

  //  rvalue = false;
  //  if(isnear(toll, primX->x1, primY->y1, primX->x2, primY->y2))
  return false;

}

double Primitive::getdata(int x, int y) {
  double localdata;
  localdata = this->data[y+3*x];
  return localdata;
}

void Primitive::setdata(int x, int y, double value) {
  this->data[y+3*x] = value;
}

int Primitive::allocatedata(int ndata) {
  if(ndata == 0){
    /* We requested to allocate zero data */
    fprintf(stdout, "Requested an alocation of zero [i|s|h]data, skipping data allocation\n");
    return 0;
  }
  else if(this->data == NULL) {
    if(this->maxdata == 0) {
      fprintf(stdout, "Allocating a new mem chunk for [i|s|h]data for " \
                      "%i elements (%i bytes): ", ndata, ndata*3*sizeof(double));
      this->data = (double *)calloc(sizeof(double), ndata * 3);
      if(this->data == NULL) { 
	fprintf(stdout, "Failed\n");
	fprintf(stderr, "[i|s|h]data allocation failed\n");
	return -1;  // memory allocation failed
      }
      else {
	fprintf(stdout, "Success\n");
	this->maxdata = ndata;
	return 0; // OK!
      }
    }
    else {
      fprintf(stderr, "WTF?! data is null but maxdata is not 0!!\n");
      return -2; // Primitive data mismatch
    }
  }
  else {
    if(ndata < (this->maxdata) * 2) { 
      /* We dont' reallocate memory if the mem avaiable is allready 
       * big enougth but less than double of avaiable. We waste but 
       * not too much
       */ 
      fprintf(stdout, "Data chunk was allready enougth big (but less" \
                      "than double), we waste mem but save cpu\n");
      return 0;
    }
    else {
      fprintf(stdout, "Reallocating old mem chunk of %i elements (%i bytes)" \
                      "to a new chunk of %i elements (%i bytes): ", \
                       this->maxdata, this->maxdata*3*sizeof(double),\
                       ndata, ndata*3*sizeof(double));
      this->data = (double *)realloc(this->data, 3*sizeof(double)*ndata);
      if(this->data == NULL) {
	fprintf(stdout, "Failed\n");
	fprintf(stderr, "[i|s|h]data reallocation failed\n");
	return -1;  // memory reallocation failed
      }
      else {
	fprintf(stdout, "Success\n");
	this->maxdata = ndata;
	return 0; // OK!
      }
    }
  }
  return -3; // How the hell we got here???!!!!
}

void Line::set_coord(int spli,\
		     double X1, double Y1,\
		     double X2, double Y2){
  spline = spli;
  x1 = X1;
  y1 = Y1;
  x2 = X2;
  y2 = Y2;
}

void Circle::find_extremes(){
  x1 = centerx + radius*cos(startarc / 360 * 2.0 * PI);
  y1 = centery + radius*sin(startarc / 360 * 2.0 * PI);
  x2 = centerx + radius*cos(endarc / 360 * 2.0 * PI);
  y2 = centery + radius*sin(endarc / 360 * 2.0 * PI);
}

void Circle::set_coord(int spli, double rad,\
		       double centx, double centy,\
		       double sarc, double earc){
  this->spline = spli;
  this->radius = rad;
  this->centerx = centx;
  this->centery = centy;
  this->startarc = sarc;
  this->endarc = earc;
  this->find_extremes();
}

void Ellipse::find_extremes(){
  x1 = centerx + semix * cos(startarc / 360 * 2.0 * PI);
  y1 = centery + semiy * sin(startarc / 360 * 2.0 * PI);
  x2 = centerx + semix * cos(endarc / 360 * 2.0 * PI);
  y2 = centery + semiy * sin(endarc / 360 * 2.0 * PI);
}

void Ellipse::set_coord(int spli,\
			double semx, double semy,\
			double centx, double centy,\
			double sarc, double earc){
  this->spline = spli;
  this->semix = semx;
  this->semiy = semy;
  this->centerx = centx;
  this->centery = centy;
  this->startarc = sarc;
  this->endarc = earc;
  this->find_extremes();
}


u2d_Core::u2d_Core(){
  firstprimitive = NULL;
  lastprimitive = NULL;
  globalprimid = 0;
  elements = 0;
  //  vertexs = 0;
}


/* PRIMITIVE LIST MANGEMENT */

Primitive *u2d_Core::new_prim(int type) {
  Primitive *prim = NULL;
  elements++;
  //  vertexs++;
  switch(type) {
  case PR_LINE:
    prim = new Line;
    break;
  case PR_CIRCLE:
    prim = new Circle;
    break;
  case PR_ELLIPSE:
    prim = new Ellipse;
    break;
  case PR_DATA:
    prim = new Data;
    break;
  default:
    elements--;
    //    vertexs--;
    fprintf(stderr, "Unknown primitive type\n");
  }
  return prim;
}

Primitive *u2d_Core::add_prim(int type) {
  Primitive *newprim;
  newprim = new_prim(type);
    if(firstprimitive == NULL){
      firstprimitive = newprim;
    }
    else {
      lastprimitive->nextprimitive = newprim;
    }
    newprim->prevprimitive = lastprimitive;
    lastprimitive = newprim;
    newprim->id = globalprimid++;
    return newprim;
}

void u2d_Core::del_prim(Primitive *oldprim) {
  if(oldprim == NULL){
    /* Just to be sure none pass us a NULL pointer */
    return;
  }
  else if(oldprim == firstprimitive && oldprim->nextprimitive != NULL){
    /* We want to delete the firstprimitive and it's not the only one */
    firstprimitive = oldprim->nextprimitive;
  }
  else if(oldprim == firstprimitive && oldprim->nextprimitive == NULL){
    /* We want to delete the firstprimitive and it's the only one */
    firstprimitive = NULL;
    lastprimitive = NULL;
  }
  else if(oldprim->nextprimitive != NULL) {
    /* oldprim is not the first, because we checked that above
     * and is not the last because it points to anoter one. It
     * is in the middle of the list */
    oldprim->prevprimitive->nextprimitive = oldprim->nextprimitive;
  }
  else if (oldprim->nextprimitive == NULL) {
    /* oldprim should be the last, but it is not the first and only
     * because we checked that above */
    oldprim->prevprimitive->nextprimitive = NULL;
    lastprimitive = oldprim->prevprimitive;
  }
  delete oldprim;
}

int u2d_Core::get_prim_number(){
  return elements;
}

Primitive *u2d_Core::get_prim_list_start(){
  return firstprimitive;
}
Primitive *u2d_Core::get_prim_list_end(){
  return lastprimitive;
}
