#ifndef U2D_CORE_H
#define U2D_CORE_H

/* 

    umesh2dgui, a GUI for umesh2d grid generator 

    Copyright (C) 2003 Gabriele Dini Ciacci
  
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <math.h>


#define DEFAULT_TOLL 0.000001
#define PI 3.14159265358979

#define PR_LINE 1
#define PR_CIRCLE 2
#define PR_ELLIPSE 3
#define PR_DATA 4

class Primitive {
 private:
  /* find_extremes() set the two extreme points where primitive start
   * and where it ends.
   * Does notting for PR_LINE naturally.
   * Each memeber define his funcion
   */
  void find_extremes(){ };
 public:
  Primitive(){
    nextprimitive = NULL;
    prevprimitive = NULL;
    maxdata = 0;
    curdata = 0;
    data = NULL;
    object = 0;
    object_part = 0;
    isboundary = false;
  }
  ~Primitive(){  }
  int id;
  int type;
  char primname[34];
  double *data;
  /* The object the primitive is part of.
   * An object is a group of one or more primitives.
   * 0 means unknown */
  unsigned int object;
  /* Progressive number of the primitive inside his object
   * 0 means it's unassigned */
  unsigned int object_part;
  /* Say if a primitive is part of the external boundary or not */
  bool isboundary;
  /* NOTE: a data element is a triplet of [i|s|h]data */
  int maxdata; // maximum number of data elements for witch mem is allcoated
  int curdata; // current number of data elements memorized
  double getdata(int x, int y); // y from 0 to 2
  void setdata(int x, int y, double value); // y from 0 to 2
  int allocatedata(int ndata);
  class Primitive *nextprimitive;
  class Primitive *prevprimitive;
  void set_coord(){ };
};

class Line : public Primitive {
 public:
  Line() {
    type = PR_LINE;
  }
  ~Line(){  }
  int spline;
  double x1;
  double y1;
  double x2;
  double y2;
  void set_coord(int spli,\
		 double X1, double Y1,\
		 double X2, double Y2);
};

class Circle : public Primitive {
 public:
  Circle() {
    type = PR_CIRCLE;
  }
  ~Circle() {  }
  int spline;
  double radius;
  double centerx;
  double centery;
  double startarc;
  double endarc;
  /* start and end point of the curve */
  double x1;
  double y1;
  double x2;
  double y2;
  void set_coord(int spli, double rad,\
		 double centx, double centy,\
		 double sarc, double earc);
 private:
  void find_extremes();
};

class Ellipse : public Primitive {
 public:
  Ellipse() {
    type = PR_ELLIPSE;
  }
  ~Ellipse() {  }
  int spline;
  double semix;
  double semiy;
  double centerx;
  double centery;
  double startarc;
  double endarc;
  /* start and end point of the curve */
  double x1;
  double y1;
  double x2;
  double y2;
  void set_coord(int spli,\
		 double semx, double semy,\
		 double centx, double centy,\
		 double sarc, double earc);
 private:
  void find_extremes();
};

class Data : public Primitive {
};


/* Functions to find if a point is near another.
 * We use those to actualy consider really near points as the same point
 * The one with primitive arguments are not impelented yet
 */
bool isnear(double toll, double x1, double y1);
bool isnear(double toll, double x1, double x2, double y1, double y2);
bool isnear(double toll, Primitive *primX, Primitive *primY);


class u2d_Core {
 public:
  u2d_Core();
  ~u2d_Core() {  }

  /* PRIMITIVE MANGEMENT */
  Primitive *add_prim(int type);
  void del_prim(class Primitive *oldprim);
  int get_prim_number();
  Primitive *get_prim_list_start();
  Primitive *get_prim_list_end();

  /* Assign objects numbers to primitives in the main list */
  void find_objects(Primitive *prim);

 private:
  /* PRIMITIVE MANGEMENT */
  Primitive *firstprimitive; 
  Primitive *lastprimitive;
  Primitive *new_prim(int type);
  int elements;
  int globalprimid; /* Consecutive number assigned to any
		     * generated primitive, can overflow */
  // in the future the id of the prim need to be managed much better.

};

#endif
